@@ -1,4 +1,6 @@
+from gnwmanager.cli._bootloader import flash_bootloader
 from gnwmanager.cli._debug import debug
+from gnwmanager.cli._dump import dump
 from gnwmanager.cli._erase import erase
 from gnwmanager.cli._filesystem import (
     format,
@@ -13,8 +15,10 @@ from gnwmanager.cli._gdb import gdb, gdbserver
 from gnwmanager.cli._install import install
 from gnwmanager.cli._lock import lock
 from gnwmanager.cli._monitor import monitor
+from gnwmanager.cli._patch import flash_patch
 from gnwmanager.cli._pull import pull
 from gnwmanager.cli._push import push
 from gnwmanager.cli._screenshot import screenshot
+from gnwmanager.cli._sdcard import sdpush
 from gnwmanager.cli._start import start
 from gnwmanager.cli._unlock import unlock
