@@ -2,6 +2,7 @@ import importlib.resources
 import logging
 from collections import namedtuple
 from copy import deepcopy
+from itertools import count
 from math import ceil
 from time import sleep, time
 from typing import Dict, List, Literal, NamedTuple, Optional, Union
@@ -23,16 +24,19 @@ class Variable(NamedTuple):
     size: int
 
 
-actions: Dict[str, int] = {
+ERROR_MASK = 0xFFFF_0000
+
+actions: dict[str, int] = {
     "ERASE_AND_FLASH": 0,
     "HASH": 1,
+    "WRITE_FILE_TO_SD": 2,
 }
 
-_comm: Dict[str, Variable] = {
+_comm: dict[str, Variable] = {
     "framebuffer": Variable(0x2400_0000, 320 * 240 * 2),
     "flashapp_comm": Variable(0x2402_5800, 0xC4000),
 }
-_contexts: List[Dict[str, Variable]] = [{} for _ in range(2)]
+_contexts: list[dict[str, Variable]] = [{} for _ in range(2)]
 
 
 def _populate_comm():
@@ -45,6 +49,9 @@ def _populate_comm():
     _comm["min_erase_size"] = last_variable = Variable(last_variable.address + last_variable.size, 4)
     _comm["upload_in_progress"] = last_variable = Variable(last_variable.address + last_variable.size, 4)
     _comm["download_in_progress"] = last_variable = Variable(last_variable.address + last_variable.size, 4)
+    _comm["expected_hash"] = last_variable = Variable(last_variable.address + last_variable.size, 32)
+    _comm["actual_hash"] = last_variable = Variable(last_variable.address + last_variable.size, 32)
+    _comm["dest_path"] = last_variable = Variable(last_variable.address + last_variable.size, 256)
 
     for i in range(2):
         struct_start = _comm["flashapp_comm"].address + ((i + 1) * 1024)
@@ -60,6 +67,9 @@ def _populate_comm():
         _contexts[i]["bank"] = last_variable = Variable(last_variable.address + last_variable.size, 4)
         _contexts[i]["action"] = last_variable = Variable(last_variable.address + last_variable.size, 4)
         _contexts[i]["response_ready"] = last_variable = Variable(last_variable.address + last_variable.size, 4)
+        _contexts[i]["dest_path"] = last_variable = Variable(last_variable.address + last_variable.size, 256)
+        _contexts[i]["block"] = last_variable = Variable(last_variable.address + last_variable.size, 4)
+        _contexts[i]["total_blocks"] = last_variable = Variable(last_variable.address + last_variable.size, 4)
 
         _contexts[i]["ready"] = last_variable = Variable(last_variable.address + last_variable.size, 4)
 
@@ -77,7 +87,7 @@ def _round_up(value, mod) -> int:
     return int(ceil(value / mod) * mod)
 
 
-def _chunk_bytes(data: bytes, chunk_size: int) -> List[bytes]:
+def _chunk_bytes(data: bytes, chunk_size: int) -> list[bytes]:
     return [data[i : i + chunk_size] for i in range(0, len(data), chunk_size)]
 
 
@@ -151,24 +161,26 @@ class GnW:
         addr = _key_to_address(key)
         self.backend.write_memory(addr, val)
 
+    def write_str(self, key: Union[int, str, Variable], val: str):
+        byte_val = val.encode("utf-8") + b"\x00"
+        self.write_memory(key, byte_val)
+
     def wait_for_idle(self, timeout: float = 20):
         """Block until the on-device status is IDLE."""
         log.debug("Waiting for device to idle.")
         t_start = time()
         t_deadline = t_start + timeout
-        error_mask = 0xFFFF_0000
 
-        while True:
-            status_enum = self.read_uint32("status")
-            status_str = flashapp_status_enum_to_str.get(status_enum, "UNKNOWN")
+        for i in count():
+            status_str = self._get_status()
             if status_str == "IDLE":
                 break
-            elif (status_enum & error_mask) == 0xBAD0_0000:
-                raise DataError(status_str)
+            if i % 10 == 0:
+                log.debug(f"waiting for device to IDLE; current status: {status_str}")
             if time() > t_deadline:
                 raise TimeoutError("wait_for_idle")
-            sleep(0.05)
-        log.debug(f"Waited {time() - t_start:.3}s for device idle.")
+            sleep(0.1)
+        log.debug(f"Waited {time() - t_start:.3f}s for device idle.")
 
     def wait_for_all_contexts_complete(self, timeout=20):
         log.debug("Waiting for all contexts to complete.")
@@ -178,9 +190,9 @@ class GnW:
             while self.read_uint32(context["ready"]):
                 if time() > t_deadline:
                     raise TimeoutError("wait_for_all_contexts_complete")
-                sleep(0.05)
+                sleep(0.1)
             log.debug(f"Context {i} complete.")
-        log.debug(f"Waited {time() - t_start:.3}s for all contexts to complete.")
+        log.debug(f"Waited {time() - t_start:.3f}s for all contexts to complete.")
         self.wait_for_idle(timeout=t_deadline - time())
 
     def wait_for_context_response(self, context, timeout=20):
@@ -191,8 +203,8 @@ class GnW:
         while not self.read_uint32(context["response_ready"]):
             if time() > t_deadline:
                 raise TimeoutError("wait_for_context_response")
-            sleep(0.05)
-        log.debug(f"Waited {time() - t_start:.3}s for context {context_index} response.")
+            sleep(0.1)
+        log.debug(f"Waited {time() - t_start:.3f}s for context {context_index} response.")
 
     def reset_context_counter(self):
         self.context_counter = 1
@@ -210,18 +222,31 @@ class GnW:
         self.reset_context_counter()
         self._gnwmanager_started = False
 
+    def _get_status(self, raise_on_error=True) -> str:
+        status_enum = self.read_uint32("status")
+        status_str = flashapp_status_enum_to_str.get(status_enum, "UNKNOWN")
+        if raise_on_error and (status_enum & ERROR_MASK) == 0xBAD0_0000:
+            if status_str in ("BAD_HASH_RAM", "BAD_HASH_FLASH"):
+                expected_hash = self.read_memory("expected_hash")
+                actual_hash = self.read_memory("actual_hash")
+                raise DataError(f"{status_str}:\nExpected: {expected_hash.hex()}\nActual: {actual_hash.hex()}")
+            else:
+                raise DataError(status_str)
+        return status_str
+
     def get_context(self, timeout=20):
         t_start = time()
         t_deadline = t_start + timeout
         while True:
             for i, context in enumerate(_contexts):
                 if not self.read_uint32(context["ready"]):
-                    log.debug(f"Got context {i} in {time() - t_start:.3}s.")
+                    log.debug(f"Got context {i} in {time() - t_start:.3f}s.")
                     return context
+                self._get_status()
                 if time() > t_deadline:
                     log.debug(f"Timeout ({timeout}s) reached waiting to get an available context.")
                     raise TimeoutError
-            sleep(0.05)
+            sleep(0.1)
 
     def filesystem(self, offset: Optional[int] = None, **kwargs):
         from gnwmanager.filesystem import get_filesystem
@@ -232,7 +257,7 @@ class GnW:
 
         return get_filesystem(self, offset=offset, **kwargs)
 
-    def read_hashes(self, offset, size) -> List[bytes]:
+    def read_hashes(self, offset, size) -> list[bytes]:
         """Blocking call to get the hashes of external flash chunks.
 
         All chunks are 256KB; the last chunk may be less.
@@ -343,7 +368,8 @@ class GnW:
         else:
             self.write_uint32(context["erase"], 0)
 
-        self.write_memory(context["expected_sha256"], sha256(data))
+        data_hash = sha256(data)
+        self.write_memory(context["expected_sha256"], data_hash)
 
         if compress:
             self.write_uint32(context["compressed_size"], len(compressed_data))
@@ -352,6 +378,7 @@ class GnW:
             self.write_uint32(context["compressed_size"], 0)
             self.write_memory(context["buffer"], data)
 
+        log.debug(f"Activating PROGRAM: {data_hash.hex()}")
         self.write_uint32(context["ready"], self.context_counter)
         self.context_counter += 1
         log.debug(f"context_counter incremented to {self.context_counter}.")
@@ -468,8 +495,6 @@ class GnW:
 
         chunk_size = self.contexts[0]["buffer"].size  # Assumes all contexts have same size buffer
         chunks = chunk_bytes(data, chunk_size)
-        len(chunks)
-        [sha256(chunk) for chunk in chunks]
 
         Packet = namedtuple("Packet", ["addr", "data"])
         all_packets = [Packet(offset + i * chunk_size, chunk) for i, chunk in enumerate(chunks)]
@@ -488,17 +513,110 @@ class GnW:
 
         self.wait_for_all_contexts_complete()
 
-    def start_gnwmanager(self, force=False):
+    def _sd_write_file_chunk(
+        self,
+        path: str,
+        block: int,
+        total_blocks: int,
+        data: bytes = b"",
+        blocking: bool = True,
+        compress: bool = True,
+    ) -> None:
+        """Low-level write data to fat filesystem.
+
+        Limited to RAM constraints (i.e. <256KB writes).
+
+        Parameters
+        ----------
+        path: str
+            Path of file to write
+        size: int
+            Number of bytes to write.
+        blocking: bool
+            Wait for action to be complete.
+        """
+        log.debug(f"gnw._sd_write_file_chunk: {path=} {len(data)=} {blocking=} {compress=}")
+
+        if not path:
+            raise ValueError("Destination SD path cannot be empty.")
+
+        if len(data) > (256 << 10):
+            raise ValueError("Too large of data for a single write.")
+
+        if compress:
+            compressed_data = compress_lzma(data)
+            # If we are unable to compress meaningfully, don't bother.
+            if len(compressed_data) > (0.9 * len(data)):
+                compress = False
+        else:
+            compressed_data = b""
+
+        context = self.get_context()
+
+        log.debug("setting upload_in_progress.")
+        self.write_uint32("upload_in_progress", 1)
+
+        self.write_uint32(context["action"], actions["WRITE_FILE_TO_SD"])
+        self.write_str(context["dest_path"], path)
+        self.write_uint32(context["size"], len(data))
+        self.write_uint32(context["block"], block)
+        self.write_uint32(context["total_blocks"], total_blocks)
+
+        data_hash = sha256(data)
+        self.write_memory(context["expected_sha256"], data_hash)
+
+        if compress:
+            self.write_uint32(context["compressed_size"], len(compressed_data))
+            self.write_memory(context["buffer"], compressed_data)
+        else:
+            self.write_uint32(context["compressed_size"], 0)
+            self.write_memory(context["buffer"], data)
+
+        log.debug(f"Activating PROGRAM: {data_hash.hex()}")
+        self.write_uint32(context["ready"], self.context_counter)
+        self.context_counter += 1
+        log.debug(f"context_counter incremented to {self.context_counter}.")
+
+        log.debug("clearing upload_in_progress.")
+        self.write_uint32("upload_in_progress", 0)
+
+        if blocking:
+            self.wait_for_all_contexts_complete()
+
+    def sd_write_file(
+        self,
+        path: str,
+        data: bytes,
+        progress: bool = False,
+    ):
+        if path.endswith("/"):
+            raise ValueError(f"path shall not be a folder {path}")
+        if not path.startswith("/"):
+            raise ValueError(f"path shall start with '/' {path}")
+        chunk_size = self.contexts[0]["buffer"].size  # Assumes all contexts have same size buffer
+        chunks = chunk_bytes(data, chunk_size)
+
+        log.info(f"Data chunked into {len(chunks)} packets.")
+
+        if len(chunks) == 0:
+            log.info("Programming empty file.")
+            self._sd_write_file_chunk(path, 0, 1, blocking=False)
+
+        for i, packet in enumerate(tqdm(chunks) if progress else chunks):
+            log.info(f"Programming packet {i + 1}/{len(chunks)}.")
+            self._sd_write_file_chunk(path, i, len(chunks), packet, blocking=False)
+            self.write_uint32("progress", int(26 * (i + 1) / len(chunks)))
+
+        self.wait_for_all_contexts_complete()
+
+    def start_gnwmanager(self, force=False, resume=True):
         if not force and self._gnwmanager_started:
             return
 
         self.reset_and_halt()
 
-        # TODO: this is deprecated, but the replacement was introduced in python3.9.
-        # Migrate to ``as_file`` once python3.8 hits EOL.
-        with importlib.resources.path("gnwmanager", "firmware.bin") as f:
-            firmware = f.read_bytes()
-            log.debug(f"Loaded {len(firmware)} bytes of gnwmanager firmware.")
+        firmware = (importlib.resources.files("gnwmanager") / "firmware.bin").read_bytes()
+        log.debug(f"Loaded {len(firmware)} bytes of gnwmanager firmware.")
 
         log.debug("Loading gnwmanager firmware to device.")
         self.write_memory(0x240E_6800, firmware)  # See STM32H7B0VBTx_FLASH.ld
@@ -513,11 +631,14 @@ class GnW:
         self.backend.write_register("pc", pc)
 
         log.debug("Resuming chip execution.")
-        self.backend.resume()
-        self.wait_for_idle()
 
-        log.debug("Setting device time.")
-        self.write_uint32("utc_timestamp", timestamp_now())
+        if resume:
+            self.backend.resume()
+            self.wait_for_idle()
+
+            # Time has to be set **after** device is in an idle state
+            log.debug("Setting device time.")
+            self.write_uint32("utc_timestamp", timestamp_now())
 
         self._gnwmanager_started = True
 
